@@ -117,8 +117,9 @@ mhs_history_class_init (MhsHistoryClass *klass)
                   G_SIGNAL_RUN_LAST,
                   G_STRUCT_OFFSET (MhsHistoryClass, pinned_page),
                   NULL, NULL,
-                  mhs_marshal_VOID__STRING_STRING_INT,
-                  G_TYPE_NONE, 3, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_INT);
+                  mhs_marshal_VOID__STRING_STRING_INT_BOOL,
+                  G_TYPE_NONE, 4,
+                  G_TYPE_STRING, G_TYPE_STRING, G_TYPE_INT, G_TYPE_BOOLEAN);
 
   history_signals[UNPINNED_PAGE_SIGNAL] =
     g_signal_new ("unpinned-page",
@@ -147,11 +148,12 @@ mhs_history_class_init (MhsHistoryClass *klass)
                                      G_TYPE_STRING,
                                      G_TYPE_INT,
                                      G_TYPE_INVALID);
-  dbus_g_object_register_marshaller (mhs_marshal_VOID__STRING_STRING_INT,
+  dbus_g_object_register_marshaller (mhs_marshal_VOID__STRING_STRING_INT_BOOL,
                                      G_TYPE_NONE,
                                      G_TYPE_STRING,
                                      G_TYPE_STRING,
                                      G_TYPE_INT,
+                                     G_TYPE_BOOLEAN,
                                      G_TYPE_INVALID);
 }
 
@@ -191,10 +193,11 @@ mhs_history_pinned_page_cb (DBusGProxy  *proxy,
                             gchar *title,
                             gchar *url,
                             glong visit_time,
+                            gboolean more_pending,
                             gpointer data)
 {
   g_signal_emit (data, history_signals[PINNED_PAGE_SIGNAL], 0,
-                 title, url, visit_time);
+                 title, url, visit_time, more_pending);
 }
 
 static void
@@ -246,6 +249,7 @@ mhs_history_init (MhsHistory *self)
                                G_TYPE_STRING,
                                G_TYPE_STRING,
                                G_TYPE_INT,
+                               G_TYPE_BOOLEAN,
                                G_TYPE_INVALID);
       dbus_g_proxy_add_signal (priv->proxy, "UnpinnedPage",
                                G_TYPE_STRING,
@@ -626,6 +630,31 @@ mhs_history_get_favorites (MhsHistory *self)
                                               self);
 }
 
+gboolean
+mhs_history_remove_favorite (MhsHistory *self,
+                             const gchar *uri,
+                             GError **error)
+{
+  MhsHistoryPrivate *priv;
+  gboolean ret;
+
+  g_return_val_if_fail (MHS_IS_HISTORY (self), FALSE);
+  g_return_val_if_fail (error == NULL || *error == NULL, FALSE);
+
+  priv = self->priv;
+
+  if (!mhs_history_check_proxy (self, error))
+    return FALSE;
+
+  ret = org_moblin_mhs_History_remove_favorite (priv->proxy,
+                                                uri,
+                                                error);
+
+  _mhs_error_translate_from_dbus (error);
+
+  return ret;
+}
+
 gboolean
 mhs_history_clear_history (MhsHistory *self,
                            GError **error)
