@@ -93,6 +93,10 @@ static gboolean mhs_history_get_favicon (MhsHistory *self,
 static gboolean mhs_history_get_favorites (MhsHistory *self,
                                            GError **error);
 
+static gboolean mhs_history_remove_favorite (MhsHistory *self,
+                                             const gchar *uri,
+                                             GError **error);
+
 static gboolean mhs_history_clear_history (MhsHistory *self,
                                            GError **error);
 
@@ -465,7 +469,8 @@ mhs_history_emit_pinned_page_for_item (MhsHistory *self,
                                        bookmark_service,
                                        nsCOMPtr<nsINavHistoryService>
                                        history_service,
-                                       PRInt64 bookmark_id)
+                                       PRInt64 bookmark_id,
+                                       gboolean more_pending)
 {
   nsresult rv = NS_OK;
   nsCAutoString title, uri_str;
@@ -540,7 +545,8 @@ mhs_history_emit_pinned_page_for_item (MhsHistory *self,
 
       g_signal_emit (self, history_signals[PINNED_PAGE_SIGNAL], 0,
                      title.get (), uri_str.get (),
-                     (gint) (visit_time / G_USEC_PER_SEC));
+                     (gint) (visit_time / G_USEC_PER_SEC),
+                     more_pending);
     }
 
  rv_error:
@@ -566,7 +572,8 @@ MhsBookmarksObserver::OnItemAdded (PRInt64 item_id,
   mhs_history_emit_pinned_page_for_item (history,
                                          bookmark_service,
                                          history_service,
-                                         item_id);
+                                         item_id,
+                                         FALSE);
   if (NS_FAILED (rv))
     return rv;
 
@@ -685,8 +692,9 @@ mhs_history_class_init (MhsHistoryClass *klass)
                   G_SIGNAL_RUN_LAST,
                   G_STRUCT_OFFSET (MhsHistoryClass, pinned_page),
                   NULL, NULL,
-                  mhs_marshal_VOID__STRING_STRING_INT,
-                  G_TYPE_NONE, 3, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_INT);
+                  mhs_marshal_VOID__STRING_STRING_INT_BOOL,
+                  G_TYPE_NONE, 4,
+                  G_TYPE_STRING, G_TYPE_STRING, G_TYPE_INT, G_TYPE_BOOLEAN);
 
   history_signals[UNPINNED_PAGE_SIGNAL] =
     g_signal_new ("unpinned-page",
@@ -1278,7 +1286,10 @@ mhs_history_favorites_cb (gpointer data)
                                 ::RESULTS_AS_SITE_QUERY);
   if (NS_FAILED (rv))
     goto rv_error;
-  query_options->SetMaxResults (MHS_HISTORY_N_FAVORITES);
+  /* Times two here in case we encounter errors while iterating over
+   * the result (i.e. fallback results)
+   */
+  query_options->SetMaxResults (MHS_HISTORY_N_FAVORITES * 2);
   if (NS_FAILED (rv))
     goto rv_error;
   query_options->SetQueryType (nsINavHistoryQueryOptions
@@ -1310,12 +1321,11 @@ mhs_history_favorites_cb (gpointer data)
   rv = root_node->GetChildCount (&root_child_count);
   if (NS_SUCCEEDED (rv))
     {
-      PRUint32 outer;
+      PRUint32 outer, n_found;
 
-      if (root_child_count > MHS_HISTORY_N_FAVORITES)
-        root_child_count = MHS_HISTORY_N_FAVORITES;
-
-      for (outer = 0; outer < root_child_count; outer++)
+      for (outer = 0, n_found = 0;
+           (outer < root_child_count) && (n_found < MHS_HISTORY_N_FAVORITES);
+           outer++)
         {
           nsCOMPtr<nsINavHistoryResultNode> node, best_node;
           nsCOMPtr<nsINavHistoryContainerResultNode> container_node;
@@ -1337,6 +1347,7 @@ mhs_history_favorites_cb (gpointer data)
           rv = container_node->GetChildCount (&child_count);
           if (NS_FAILED (rv))
             break;
+
           for (inner = 0; inner < child_count; inner++)
             {
               nsCOMPtr<nsINavHistoryResultNode> inner_node;
@@ -1362,25 +1373,22 @@ mhs_history_favorites_cb (gpointer data)
           if (inner < child_count)
             break;
           if (best_access_count == 0)
-            {
-              rv = NS_ERROR_OUT_OF_MEMORY;
-              break;
-            }
+            continue;
 
           rv = best_node->GetUri (url);
           if (NS_FAILED (rv))
-            break;
+            continue;
           rv = best_node->GetTitle (title);
           if (NS_FAILED (rv))
-            break;
+            continue;
 
-          urls[outer] = g_strdup (url.get ());
-          titles[outer] = g_strdup (title.get ());
+          urls[n_found] = g_strdup (url.get ());
+          titles[n_found] = g_strdup (title.get ());
+          n_found ++;
         }
 
-      if (outer == root_child_count)
-        g_signal_emit (self, history_signals[FAVORITES_RECEIVED_SIGNAL], 0,
-                       urls, titles);
+      g_signal_emit (self, history_signals[FAVORITES_RECEIVED_SIGNAL], 0,
+                     urls, titles);
     }
 
   g_strfreev (titles);
@@ -1415,6 +1423,45 @@ mhs_history_get_favorites (MhsHistory *self,
   return TRUE;
 }
 
+static gboolean
+mhs_history_remove_favorite (MhsHistory *self,
+                             const gchar *uri_str,
+                             GError **error)
+{
+  g_return_val_if_fail (MHS_IS_HISTORY (self), FALSE);
+
+  nsresult rv;
+
+  {
+    nsCOMPtr<nsIURI> uri;
+    rv = NS_NewURI (getter_AddRefs (uri), uri_str);
+    if (NS_FAILED (rv))
+      goto rv_error;
+
+    nsCOMPtr<nsIBrowserHistory> browser_history
+      = do_GetService (NS_GLOBALHISTORY2_CONTRACTID, &rv);
+    if (NS_FAILED (rv))
+      goto rv_error;
+
+    nsCAutoString host;
+    rv = uri->GetHost(host);
+    if (NS_FAILED (rv))
+      goto rv_error;
+
+    // The favorites are grouped by host so to remove this favorite we
+    // need to remove all history entries for the host of the URI
+    rv = browser_history->RemovePagesFromHost (host, PR_FALSE);
+    if (NS_FAILED (rv))
+      goto rv_error;
+  }
+
+  return TRUE;
+
+ rv_error:
+  mhs_error_set_from_nsresult (rv, error);
+  return FALSE;
+}
+
 static gboolean
 mhs_history_clear_history (MhsHistory *self,
                            GError **error)
@@ -1499,24 +1546,30 @@ mhs_history_pinned_pages_cb (gpointer data)
   if (NS_FAILED (rv))
     goto rv_error;
 
-  while (TRUE)
+  rv = bookmark_service->GetIdForItemAt (bookmarks_menu_folder, index++,
+                                         &bookmark_id);
+  if (NS_FAILED (rv))
+    goto rv_error;
+
+  while (bookmark_id != -1)
     {
       nsCOMPtr<nsIURI> uri;
+      PRInt64 next_bookmark_id;
 
       rv = bookmark_service->GetIdForItemAt (bookmarks_menu_folder, index++,
-                                             &bookmark_id);
+                                             &next_bookmark_id);
       if (NS_FAILED (rv))
         goto rv_error;
 
-      if (bookmark_id == -1)
-        break;
-
       rv = mhs_history_emit_pinned_page_for_item (self,
                                                   bookmark_service,
                                                   history_service,
-                                                  bookmark_id);
+                                                  bookmark_id,
+                                                  next_bookmark_id != -1);
       if (NS_FAILED (rv))
         goto rv_error;
+
+      bookmark_id = next_bookmark_id;
     }
 
  rv_error:
