@@ -24,12 +24,16 @@
 
 #include <signal.h>
 #include <glib.h>
+#include <dbus/dbus.h>
 #include <dbus/dbus-glib.h>
 #include <dbus/dbus-glib-bindings.h>
+#include <dbus/dbus-glib-lowlevel.h>
 #include <nsXPCOM.h>
 #include <nsCOMPtr.h>
 #include <nsServiceManagerUtils.h>
+#include <nsComponentManagerUtils.h>
 #include <nsIDirectoryService.h>
+#include <nsIObserverService.h>
 #include <nsAppDirectoryServiceDefs.h>
 #include <nsDirectoryServiceDefs.h>
 #include <nsStringAPI.h>
@@ -68,6 +72,7 @@
 #include "mhs-prefs.h"
 #include "mhs-cookies.h"
 #include "mhs-login-manager-storage.h"
+#include "mhs-permission-manager.h"
 #include "mhs-error-private.h"
 #include "mhs-event-source.h"
 #include "mhs-enum-types.h"
@@ -433,6 +438,19 @@ mhs_signal_cb (int sig)
   g_idle_add_full (G_PRIORITY_HIGH, mhs_quit_cb, NULL, NULL);
 }
 
+static void
+mhs_service_load_thebes (void)
+{
+  /* This creates a temporary Thebes image just to ensure that the
+     Thebes modules has been loaded. Otherwise when we try to
+     decompress a PNG image (such as from a favicon) Mozilla will
+     directly try to use gfxPlatform::GetPlatform() but nothing will
+     have created the platform instance and MHS will crash. See Moblin
+     bug #5012 for more information */
+  nsCOMPtr<nsISupports> image;
+  image = do_CreateInstance ("@mozilla.org/gfx/image;1");
+}
+
 static void
 mhs_service_run_main_loop (GError **error)
 {
@@ -461,6 +479,7 @@ mhs_service_run_main_loop (GError **error)
           MhsPrefs *prefs;
           MhsCookies *cookies;
           MhsLoginManagerStorage *lms;
+          MhsPermissionManager *pm;
           GSource *event_source = mhs_event_source_new ();
 
           XRE_NotifyProfile ();
@@ -473,6 +492,7 @@ mhs_service_run_main_loop (GError **error)
           prefs = mhs_prefs_new ();
           cookies = mhs_cookies_new ();
           lms = mhs_login_manager_storage_new ();
+          pm = mhs_permission_manager_new ();
 
 #ifdef SUPPORT_GNOME_PROXY
           GConfClient *client = gconf_client_get_default ();
@@ -501,14 +521,29 @@ mhs_service_run_main_loop (GError **error)
 
           g_source_attach (event_source, NULL);
 
+          mhs_service_load_thebes ();
+
           signal (SIGINT, mhs_signal_cb);
           signal (SIGTERM, mhs_signal_cb);
           g_main_loop_run (main_loop);
 
+          /* Let the services know the application is shutting down */
+          nsCOMPtr<nsIObserverService> observer_service
+            = do_GetService ("@mozilla.org/observer-service;1", &rv);
+          if (NS_SUCCEEDED (rv))
+            {
+              NS_NAMED_LITERAL_STRING (shutdown_str, "shutdown");
+              observer_service->NotifyObservers (nsnull,
+                                                 "quit-application",
+                                                 shutdown_str.get ());
+            }
+          observer_service = nsnull;
+
           g_object_unref (G_OBJECT (prefs));
           g_object_unref (G_OBJECT (history));
           g_object_unref (G_OBJECT (cookies));
           g_object_unref (G_OBJECT (lms));
+          g_object_unref (G_OBJECT (pm));
 
           g_source_destroy (event_source);
           g_source_unref (event_source);
@@ -547,6 +582,23 @@ mhs_register_dbus_object (DBusGProxy   *proxy,
   return FALSE;
 }
 
+static DBusHandlerResult
+mhs_dbus_filter (DBusConnection *connection,
+                 DBusMessage *message,
+                 void *user_data)
+{
+  if (dbus_message_is_signal (message,
+                              DBUS_INTERFACE_LOCAL,
+                              "Disconnected"))
+    {
+      GMainLoop *main_loop_data = (GMainLoop *) user_data;
+      g_main_loop_quit (main_loop_data);
+      return DBUS_HANDLER_RESULT_HANDLED;
+    }
+
+  return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
+}
+
 int
 main (int argc, char **argv)
 {
@@ -567,10 +619,21 @@ main (int argc, char **argv)
     }
   else
     {
-      DBusGProxy *proxy = dbus_g_proxy_new_for_name (bus,
-                                                     DBUS_SERVICE_DBUS,
-                                                     DBUS_PATH_DBUS,
-                                                     DBUS_INTERFACE_DBUS);
+      DBusGProxy *proxy;
+      DBusConnection *dcon;
+
+      /* Don't exit on disconnect */
+      dcon = dbus_g_connection_get_connection (bus);
+      dbus_connection_set_exit_on_disconnect (dcon, FALSE);
+      /* Install a message filter to handle the disconnect signal */
+      dbus_connection_add_filter (dcon, mhs_dbus_filter,
+                                  g_main_loop_ref (main_loop),
+                                  (DBusFreeFunction) g_main_loop_unref);
+
+      proxy = dbus_g_proxy_new_for_name (bus,
+                                         DBUS_SERVICE_DBUS,
+                                         DBUS_PATH_DBUS,
+                                         DBUS_INTERFACE_DBUS);
 
       if (!mhs_register_dbus_object (proxy, MHS_SERVICE_HISTORY))
         ret = MHS_SERVICE_EXIT_ERROR;
@@ -581,6 +644,9 @@ main (int argc, char **argv)
       else if (!mhs_register_dbus_object (proxy,
                                           MHS_SERVICE_LOGIN_MANAGER_STORAGE))
         ret = MHS_SERVICE_EXIT_ERROR;
+      else if (!mhs_register_dbus_object (proxy,
+                                          MHS_SERVICE_PERMISSION_MANAGER))
+        ret = MHS_SERVICE_EXIT_ERROR;
       else
         {
           dbus_g_error_domain_register (MHS_ERROR, NULL, MHS_TYPE_ERROR);
